# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from packaging.version import Version


__all__ = ['normalize_version']


def normalize_version(version: str) -> Version:
    # Remove prefixes and replace characters
    # Examples:
    #   v24.4_r1ю0ю0 -> 24.4.1.0.0
    #   v2alpha1 -> 2.1
    #   v3-beta-4 -> 3.4
    #   1.0.0 -> 1.0.0
    version = re.sub(r'[a-zA-Z]+', '.', version.lstrip('v'))
    version = re.sub(r'[._-]+', '.', version)
    return Version(version.strip('.'))
